# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

package MyR5Dragon;
use 5.010;
use strict;
use warnings;
use Carp 'croak';
use List::Util 'sum';
use Math::BaseCnv 'cnv';
use Math::Complex 'pi';
use Math::PlanePath::Base::Digits
  'digit_split_lowtohigh', 'digit_join_lowtohigh';
use Memoize;

use base 'Exporter';
our @EXPORT_OK = ('points_str','points_str_gp',
                  'chamfer','xy_to_node_name',
                  'is_integer','is_pow2','to_bigint','bigint_class',
                  'line_is_horizontal','line_is_diagonal','line_sort_ends',
                  '@dir4_to_dx','@dir4_to_dy','dxdy_to_dir4',
                  '@digit_to_x','@digit_to_y','@digit_to_dir',
                  'xy_rotate_plus90','xy_rotate_minus90',
                  'mul','div','unfold',
                  'rect_corners_sort',
                  'hull_expand','hull_and_end','hull_points_rotate_first_horizontal',
                  'quad_expand',
                  'path_tree_depth_to_hull_area',
                  'path_tree_depth_to_row_hull_area',
                  'hull_and_end',
                  'gcd',
                  'turn','LowestNon0', 'TurnRuns3',
                  'TurnLpred_flat',  'TurnRpred_flat',
                  'TurnLLpred_flat', 'TurnRRpred_flat',
                  'TurnLLLpred_flat','TurnRRRpred_flat',
                  'dir','dir4_flat','@dir4_state_table',
                  'A','AR','J','JN','JNother','GJ_xy',

                  'Rpred','@Rpred_state_table','@Rpred_disallowed_pairs',
                  'Rpred_flat','NonRpred_flat',
                  'Lpred_k','@Lpred_state_table','@Lpred_disallowed_pairs',
                  'Lpred_k_flat','NonLpred_k_flat',
                  'Bpred_k_flat','NonBpred_k_flat',

                  'HQ','HQR','HQRpred_flat',

                  'other_arm_and_n','Opred_flat',
                  'Dpred_k_flat','DpredInf_flat',
                  'Spred_k_flat','SpredInf_flat',
                  'Dpred_k_LtoH_flat','DpredInf_LtoH_flat',
                  'Spred_k_LtoH_flat','SpredInf_LtoH_flat',
                  'DpredRight_k_flat','DpredLeft_k_flat',
                  'SpredRight_k_flat','SpredLeft_k_flat',

                  'DpredFirst_flat','DpredFirst_LtoH_flat',
                  'DpredSecond_flat','DpredSecond_LtoH_flat',
                  'EpredR_flat','EpredR_LtoH_flat',
                  'EpredL_flat','EpredL_LtoH_flat',
                  'Epred_flat','EpredR3_flat','EpredL3_flat','EpredSix_flat',

                  '$R_square_segs','$L_square_segs','$B_squares_segs',

                  '$hullstouch_segs',
                  'fBpred_hullstouch_init',
                  '@fBpred_hullstouch_state_table','@fBpred_hullstouch_state_enclosed_lists',
                  '@fBpred_hullstouch_state_to_conf',
                  '$fNonRpred_hullstouch_flat','$fNonLpred_hullstouch_flat',
                  '$fNonBpred_hullstouch_flat',
                  '$fRpred_hullstouch_flat','$fLpred_hullstouch_flat',
                  '$fBpred_hullstouch_flat',

                  '$hullsaround_segs_outer','$hullsaround_segs_inner',
                  '$hullsaround_segs_all','$hullsaround_segs_eventually',
                  'fBpred_hullsaround_init',
                  '@fBpred_hullsaround_state_table','@fBpred_hullsaround_state_enclosed_lists',
                  '@fBpred_hullsaround_state_to_conf',
                  '$fNonRpred_hullsaround_flat','$fNonLpred_hullsaround_flat',
                  '$fNonBpred_hullsaround_flat',
                  '$fRpred_hullsaround_flat','$fLpred_hullsaround_flat',
                  '$fBpred_hullsaround_flat',

                  'Hwp','Hep','Hwp_factor',
                  'Hwpf_imag','Hepf_real',
                  'Hwpf_box_points','Hwpf_k2_hull_points',

                  'r5dragon_Graph',

                  'r5quad_area_tree_Graph','QAdiameter','QAdiameterCount',
                  'QAdomnum',

                  'r5join_area_tree_Graph','JAdiameter','JAdiameterCount',
                  'JAdomnum',

                  'QAdiampart_Graph',

                  'r5quad_turn_Graph','QTdiameter',
                 );
our %EXPORT_TAGS = ( all => \@EXPORT_OK );

# uncomment this to run the ### lines
# use Smart::Comments;


use Math::PlanePath::GcdRationals;
*gcd = \&Math::PlanePath::GcdRationals::_gcd;


#------------------------------------------------------------------------------

sub to_bigint {
  my ($n) = @_;
  return bigint_class()->new($n);
}
use constant::defer bigint_class => sub {
  require Math::BigInt;
  eval { Math::BigInt->import (try => 'GMP') };
  return "Math::BigInt";
};

use constant::defer R5DragonCurve_path => sub {
  require Math::PlanePath::R5DragonCurve;
  return Math::PlanePath::R5DragonCurve->new;
};

#------------------------------------------------------------------------------

our @dir4_to_dx = (1,0,-1,0);
our @dir4_to_dy = (0,1,0,-1);

sub dxdy_to_dir4 {
  my ($dx, $dy) = @_;
  if ($dx > 0) { return 0; }  # east
  if ($dx < 0) { return 2; }  # west
  if ($dy > 0) { return 1; }  # north
  if ($dy < 0) { return 3; }  # south
}

# return $x,$y multiplied by 1+2*i
# GP-Test  (1+2*I)*(x+I*y) == (x-2*y) + (2*x+y)*I
sub mul {
  my ($x,$y) = @_;
  return (  $x - 2*$y,
          2*$x +   $y);
}
sub bk {
  my ($k) = @_;
  my $y = 0 * $k;  # inherit bigint
  my $x = $y + 1;
  foreach (1 .. $k) { ($x,$y) = mul($x,$y); }
  return ($x, $y);
}

# GP-Test  (1-2*I)*(x+I*y) == (x+2*y) + (-2*x+y)*I
sub div {
  my ($x,$y) = @_;
  ($x,$y) = (   $x + 2*$y,
             -2*$x +   $y);
  if (($x%5) || ($y%5)) { croak "div() not multiple of 5"; }
  return ($x/5, $y/5);
}

#------------------------------------------------------------------------------
# turn() and dir()

sub LowestNon0 {
  my ($n) = @_;
  $n>=1 or croak "LowestNon0() only on n>=1";
  while ($n % 5 == 0) { $n /= 5; }
  return $n % 5;
}
sub turn {
  my ($n) = @_;
  return (LowestNon0($n)<=2 ? 1 : -1);
}

# return a FLAT DFA which matches base-5 digit strings of n with a left turn
use constant::defer TurnLpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('(0|1|2|3|4)* (1|2) 0*')  # 1,2 lowest non-0
    ->as_dfa
    ->MyFLAT::minimize
    ->MyFLAT::set_name('turn left');
};
# return a FLAT DFA which matches base-5 digit strings of n with a right turn
use constant::defer TurnRpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('(0|1|2|3|4)* (3|4) 0*') # 3,4 lowest non-0
    ->as_dfa
    ->MyFLAT::minimize
    ->MyFLAT::set_name('turn right');
};

# return a FLAT DFA which matches base-5 digit strings of n where both n and
# n+1 are left turns
use constant::defer TurnLLpred_flat => sub {
  my $f = TurnLpred_flat();
  return $f->intersect($f->MyFLAT::digits_increment (add => -1))
    ->MyFLAT::minimize
    ->MyFLAT::set_name('turn LL');
};
# return a FLAT DFA which matches base-5 digit strings of n where both n and
# n+1 are right turns
use constant::defer TurnRRpred_flat => sub {
  my $f = TurnRpred_flat();
  return $f->intersect($f->MyFLAT::digits_increment (add => -1))
    ->MyFLAT::minimize
    ->MyFLAT::set_name('turn RR');
};

# return a FLAT DFA which matches base-5 digit strings of n where all three
# of n, n+1 and n+2 are left turns
use constant::defer TurnLLLpred_flat => sub {
  return TurnLpred_flat()
    ->intersect(TurnLLpred_flat()->MyFLAT::digits_increment (add => -1))
    ->MyFLAT::minimize
    ->MyFLAT::set_name('turn LLL');
};
# return a FLAT DFA which matches base-5 digit strings of n where all three
# of n, n+1 and n+2 are right turns
use constant::defer TurnRRRpred_flat => sub {
  return TurnRpred_flat()
    ->intersect(TurnRRpred_flat()->MyFLAT::digits_increment (add => -1))
    ->MyFLAT::minimize
    ->MyFLAT::set_name('turn RRR');
};


our @digit_to_dir = (0,1,2,1,0);
sub dir {
  my ($n) = @_;
  my $dir = 0;
  while ($n) {
    $dir += $digit_to_dir[$n%5];
    $n = int($n/5);
  }
  return $dir;
}

our @dir4_state_table = ([0,1,2,1,0],   # 0
                         [1,2,3,2,1],   # 1
                         [2,3,0,3,2],   # 2
                         [3,0,1,0,3]);  # 3
sub dir4_flat {
  my ($want_dir4) = @_;
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@dir4_state_table,
                           name => "dir $want_dir4 mod 4",
                           accepting => $want_dir4);
}
BEGIN { memoize('dir4_flat', LIST_CACHE=>'MERGE'); }

our @digit_to_x = (0,1,1,0,0);
our @digit_to_y = (0,0,1,1,2);



#------------------------------------------------------------------------------
# Math::Geometry::Planar things

# $points is an arrayref of [$x,$y] pairs.
# Return a string "x,y x,y ...".
sub points_str {
  my ($points) = @_;
  return  join(' ',map{join(',',@$_)}@$points);
}

# $points is an arrayref of [$x,$y] pairs.
sub points_str_gp {
  my ($points) = @_;
  return join(', ',map{point_str_gp($_)}@$points);
}
sub point_str_gp {
  my ($p) = @_;
  my ($x,$y) = @$p;
  my $ret = '';
  if ($x != 0 || $y == 0) {
    $ret = "$x";
  }
  if ($y) {
    if ($ret ne '' && $y > 0) {
      $ret .= '+';
    }
    if ($y == 1) {
      $ret .= 'I';
    } elsif ($y == -1) {
      $ret .= '-I';
    } else {
      $ret .= "$y*I";
    }
  }
  return $ret;
}

#------------------------------------------------------------------------------
sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

sub is_pow2 {
  my ($n) = @_;
  (($n >> 1) < $n) or die "is_pow2() infinite";

  while ($n > 1) {
    if ($n & 1) {
      return 0;
    }
    $n >>= 1;
  }
  return ($n == 1);
}

sub chamfer {
  my ($delta, $name) = @_;
  $name //= '\MyChamfer';
  if ($delta == 0) { return ''; }
  if ($delta == 1) { return "+$name"; }
  if ($delta == -1) { return "-$name"; }
  return sprintf '%+d*%s', $delta, $name;
}

sub xy_to_node_name {
  my ($x,$y) = @_;
  if ($x==int($x) && $y==int($y)) {
    if ($x < 0) { $x = '-n'.-$x; }
    if ($y < 0) { $y = 'n'.-$y; }
    return "p$x-$y";
  } else {
    return "$x,$y";
  }
}


#------------------------------------------------------------------------------
# lines

# return true if the line x1,y1 to x2,y2 is horizontal, ie. $y1==$y2
sub line_is_horizontal {
  my ($x1,$y1, $x2,$y2) = @_;
  return $y1 == $y2;
}

# return true if the line x1,y1 to x2,y2 is diagonal, ie. slope +1
sub line_is_diagonal {
  my ($x1,$y1, $x2,$y2) = @_;
  return $x1-$y1 == $x2-$y2;
}

# return (x1,y1, x2,y2) with the ends swapped if necessary to make x1<x2, or
# if vertical x1=x2 then y1<y2
sub line_sort_ends {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2 || ($x1 == $x2 && $y1 > $y2)) {
    return ($x2,$y2, $x1,$y1);
  } else {
    return ($x1,$y1, $x2,$y2);
  }
}

sub xy_rotate_plus90 {
  my ($x,$y, $multiple) = @_;
  $multiple //= 1;
  $multiple %= 4;
  foreach (1 .. $multiple) {
    ($x,$y) = (-$y,$x);  # rotate +90
  }
  return ($x,$y);
}
sub xy_rotate_minus90 {
  my ($x,$y, $multiple) = @_;
  $multiple //= 1;
  $multiple %= 4;
  foreach (1 .. $multiple) {
    ($x,$y) = ($y,-$x);  # rotate -90
  }
  return ($x,$y);
}

sub rect_corners_sort {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2) {
    ($x1,$x2) = ($x2,$x1);
  }
  if ($y1 > $y2) {
    ($y1,$y2) = ($y2,$y1);
  }
  return ($x1,$y1, $x2,$y2);
}

#------------------------------------------------------------------------------
# Convex Hull

# return $x,$y unfolded for b^k=$ex,$ey and part $digit
sub unfold {
  my ($x,$y, $ex,$ey, $digit) = @_;
  ### unfold(): "$x,$y end=$ex,$ey digit=$digit"
  ($x,$y) = xy_rotate_plus90($x,$y, $digit_to_dir[$digit]);
  $x += $ex * $digit_to_x[$digit];
  $y += $ey * $digit_to_x[$digit];

  ($ex,$ey) = xy_rotate_plus90($ex,$ey);
  $x += $ex * $digit_to_y[$digit];
  $y += $ey * $digit_to_y[$digit];
  return ($x, $y);
}

sub hull_points_rotate_first_horizontal {
  my ($points) = @_;

  until ($points->[0]->[0] < $points->[1]->[0]
         && $points->[0]->[1] == $points->[1]->[1]) {
    push @$points, shift @$points;
  }
  return $points;
}

sub hull_expand {
  my ($points, $ex,$ey, %options) = @_;
  ### hull_expand(): "end=$ex,$ey  " . points_str($points)

  # if (@$points < 3) {
  #   ### small ...
  #   $points = [ @$points,
  #               # points_midpoint($points->[0],$points->[1],1/4),
  #               points_midpoint($points->[0],$points->[1],1/2),
  #               # points_midpoint($points->[0],$points->[1],3/4),
  #               ];
  # }

  # print " % at pxy=$ex,$ey  ",points_str($points),"\n";
  $points = [ map { my $digit = $_;
                    map {[unfold (@$_, $ex,$ey, $digit) ]} @$points
                  } 0 .. 4 ];
  ### $points

  if (@$points >= 5) {
    ### convexhull2 of: points_str($points)
    require Math::Geometry::Planar;
    my $planar = Math::Geometry::Planar->new;
    $planar->points($points);
    $planar = $planar->convexhull2;
    $points = $planar->points;
  }
  unless ($options{'unsorted'}) {
    hull_points_rotate_first_horizontal($points);
  }
  ($ex,$ey) = mul($ex,$ey);
  return ($points, $ex,$ey);
}

sub hull_and_end {
  my ($k) = @_;
  my $ex = 2;
  my $ey = 0;
  my $points = [
                [0,0],
                [$ex,$ey],
               ];

  foreach (1 .. $k) {
    ($points,$ex,$ey) = hull_expand($points,$ex,$ey);
  }
  return ($points,$ex,$ey);
}

# ascending relative to direction of $sx,$sy
sub hull_point_cmp_by_angle {
  my ($a,$b, $sx,$sy) = @_;
  require Math::NumSeq::PlanePathDelta;
  my $a4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$a);
  my $b4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$b);
  if ($a->[0] == 0 && $a->[1] == 0) { $a4 = 3.5; }
  if ($b->[0] == 0 && $b->[1] == 0) { $b4 = 3.5; }
  my $o4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($sx,$sy);
  # +.1 to allow for P1 wobbling above and below X axis
  $a4 = ($a4 - $o4 + .1); while ($a4 < 0) { $a4 += 4; }
  $b4 = ($b4 - $o4 + .1); while ($b4 < 0) { $b4 += 4; }
  return ($a4 <=> $b4);
}


sub quad_expand {
  my ($x,$y, $ex,$ey) = @_;
  ### quad_expand(): "$x,$y end=$ex,$ey"
  my @ret;
  my $ox = my $oy = 0*$ex;
  foreach (0 .. 3) {
    ### at: "xy=$x,$y oxy=$ox,$oy end $ex,$ey  ret ".join(',',@ret)
    push @ret, [$x+$ox, $y+$oy];
    $ox += $ex;
    $oy += $ey;
    ($ex,$ey) = xy_rotate_plus90($ex,$ey);
    ($x,$y) = xy_rotate_plus90($x,$y);
  }
  ### @ret
  return @ret;
}

sub quadhull_expand {
  my ($points, $ex,$ey, %options) = @_;
  ($ex,$ey) = mul($ex,$ey);

  $points = [ map {quad_expand (@$_, $ex,$ey)} @$points ];
  ### $points

  if (@$points >= 5) {
    ### convexhull2 of: points_str($points)
    require Math::Geometry::Planar;
    my $planar = Math::Geometry::Planar->new;
    $planar->points($points);
    $planar = $planar->convexhull2;
    $points = $planar->points;
  }
  unless ($options{'unsorted'}) {
    hull_points_rotate_first_horizontal($points);
  }
  return ($points, $ex,$ey);
}

#------------------------------------------------------------------------------

sub path_tree_depth_to_row_hull_area {
  my ($path, $depth) = @_;
  my $points = [ ];
  my ($n_row, $n_end) = $path->tree_depth_to_n_range($depth);
  foreach my $n ($n_row .. $n_end) {
    my ($x,$y) = $path->n_to_xy($n);
    push @$points, [$x,$y];
  }
  if (@$points < 3) {
    return 0;
  }
  return points_to_hull_area($points);
}

sub path_tree_depth_to_hull_area {
  my ($path, $depth) = @_;
  my $points = [ [1,1] ];
  foreach my $n ($path->n_start .. $path->tree_depth_to_n_end($depth)) {
    my ($x,$y) = $path->n_to_xy($n);
    push @$points, [$x,$y];
  }
  return points_to_hull_area($points);
}

sub points_to_hull_area {
  my ($points) = @_;
  if (@$points < 3) {
    return 0;
  }
  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) { $planar = $planar->convexhull2; }
  return $planar->area;
}

#------------------------------------------------------------------------------
# sequences

sub A {
  my ($k) = @_;
  return (5**$k + 1)/2 - 3**$k;
}
sub AR {
  my ($k) = @_;
  return A($k)/2;
}
sub TurnRuns3 {
  my ($k) = @_;
  return ($k==0 ? 0 : 5**($k-1)-1);
}

sub J {
  my ($k) = @_;
  return (3**$k - 1)/2;
}
sub JN {
  my ($k) = @_;
  return (3*5**$k + 1)/4;
}
sub JNother {
  my ($k) = @_;
  return (7*5**$k - 3)/4;
}

# GJ(k) = (9/10 + 3/10*I)*b^k + (2/5-1/5*I)*(b^k-1)/(3^k-1);
sub GJ_xy {
  my ($k) = @_;
  require Math::BigRat;

  my $bx = Math::BigRat->new(1);
  my $by = Math::BigRat->new(0);
  foreach (1 .. $k) {
    ($bx,$by) = mul($bx,$by);
  }
  # print "b^k = $bx $by\n";

  my $x = Math::BigRat->new('9/10') * $bx - Math::BigRat->new('3/10') * $by;
  my $y = Math::BigRat->new('9/10') * $by + Math::BigRat->new('3/10') * $bx;
  # print "xy = $x $y\n";

  $bx -= 1;
  # print "b^k-1 = $bx $by\n";
  my $t = Math::BigRat->new(3)**$k - 1;
  $bx /= $t;
  $by /= $t;
  # print "(b^k-1)/(3^k-1) = $bx $by\n";
  $x += Math::BigRat->new('2/5') * $bx + Math::BigRat->new('1/5') * $by;
  $y += Math::BigRat->new('2/5') * $by - Math::BigRat->new('1/5') * $bx;

  return ($x,$y);
}

#------------------------------------------------------------------------------
# Rpred

our @Rpred_disallowed_pairs = ('22', '23', '24',
                               '30', '32', '33', '34',
                               '40');
sub Rpred {
  my ($n) = @_;
  my $str = cnv($n,10,5);
  $str =~ s/1//g;
  foreach my $pair (@Rpred_disallowed_pairs) {
    if ($str =~ /\Q$pair/) {
      return 0;
    }
  }
  return 1;
}

our @Rpred_state_table = ([0,0,1,2,3],  # 0 = R
                          [0,1,4,4,4],  # 1 = X
                          [4,2,4,4,4],  # 2 = Y
                          [4,3,1,2,3],  # 3 = Z
                          [4,4,4,4,4]); # 4 = non

use constant::defer NonRpred_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA (\@Rpred_state_table,
                                   name => 'NonRpred',
                                   accepting => 4);
};
use constant::defer Rpred_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA (\@Rpred_state_table,
                                   name => 'Rpred',
                                   accepting_list => [0..3]);
};

#------------------------------------------------------------------------------
# Lpred

our @Lpred_disallowed_pairs = ('04',
                               '10', '11', '12', '14',
                               '20', '21', '22');
sub Lpred_of_str {
  my ($str) = @_;
  $str =~ s/3//g;
  foreach my $pair (@Lpred_disallowed_pairs) {
    if ($str =~ /\Q$pair/) {
      return 0;
    }
  }
  return 1;
}
sub Lpred_k {
  my ($k,$n) = @_;
  my $str = ($n==0 ? '' : cnv($n,10,5));
  if (length($str) > $k) { croak "Lpred_k() n=$n bigger than k=$k"; }
  return Lpred_of_str(sprintf '%0*s', $k, $str);
}

# digit reversal 0 <-> 4
our @Lpred_state_table = map {[ reverse @$_ ]} @Rpred_state_table;

use constant::defer NonLpred_k_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA (\@Lpred_state_table,
                                   name => 'NonLpred k',
                                   accepting => 4);
};
use constant::defer Lpred_k_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA (\@Lpred_state_table,
                                   name => 'Lpred k',
                                   accepting_list => [0..3]);
};


#------------------------------------------------------------------------------
# Bpred

use constant::defer NonBpred_k_flat => sub {
  my $f = NonRpred_flat()->intersect(NonLpred_k_flat())->as_dfa;
  $f->{'name'} = 'NonBpred k';
  return $f;
};
use constant::defer Bpred_k_flat => sub {
  my $f = Rpred_flat()->union(Lpred_k_flat())->as_dfa;
  $f->{'name'} = 'Bpred k';
  return $f;
};


#------------------------------------------------------------------------------
# other()

my @other_state_table = ([ 0,0,0,0,0 ],
                         [ 1,3,2,5,4 ],   # 1 = start
                         [ 3,2,6,6,2 ],   # 2 = lm0
                         [ 3,3,2,6,6 ],   # 3 = lm1
                         [ 6,6,5,4,4 ],   # 4 = rm0
                         [ 5,6,6,5,4 ],   # 5 = rm1
                         [ 6,6,6,6,6 ]);  # 6 = same
my @other_state_output = ([ 0,0,0,0,0 ],
                          [ 0,2,1,4,3 ],   # start
                          [ 2,1,3,4,0 ],   # lm0
                          [ 4,1,0,2,3 ],   # lm1
                          [ 1,2,4,3,0 ],   # rm0
                          [ 4,0,1,3,2 ],   # rm1
                          [ 0,1,2,3,4 ]);  # 6 = same
sub other_arm_and_n {
  my ($n) = @_;
  my @n_digits = (digit_split_lowtohigh($n,5), 0);
  my $state = 1;
  foreach my $i (0 .. $#n_digits) {  # low to high
    ### at: "i=$i state=$state digit=$n_digits[$i]  output=$other_state_output[$state][$n_digits[$i]]"
    my $digit = $n_digits[$i];
    $n_digits[$i] = $other_state_output[$state][$digit];
    $state = $other_state_table[$state][$digit];
  }
  ### final: "state=$state n_digits=".join(',',reverse @n_digits)
  my $other_n = digit_join_lowtohigh(\@n_digits,5);
  my $arm;
  if ($state == 6) {  # same
    $arm = 0;
  } elsif ($state == 3) {  # lm1
    $arm = -1;
    $other_n = 5**scalar(@n_digits) - $other_n;
  } elsif ($state == 5) {  # rm1
    $arm = 1;
    $other_n = 5**scalar(@n_digits) - $other_n;
  } else {
    die "oops, final state ",$state;
  }
  return ($arm, $other_n);
}

use constant::defer Dpred_k_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA (\@other_state_table,
                                   name => 'Dpred k LtoH',
                                   starting => 1,
                                   accepting => 6); # same
};
use constant::defer DpredInf_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA (\@other_state_table,
                                   name => 'DpredInf LtoH',
                                   starting => 1,
                                   accepting_list => [4,6]); # rm0, same
};

use constant::defer Spred_k_LtoH_flat => sub {
  require MyFLAT;
  return Dpred_k_LtoH_flat()->complement
    ->MyFLAT::set_name('Spred k LtoH');
};
use constant::defer SpredInf_LtoH_flat => sub {
  require MyFLAT;
  return DpredInf_LtoH_flat()->complement
    ->MyFLAT::set_name('SpredInf LtoH');
};

use constant::defer Dpred_k_flat => sub {
  require MyFLAT;
  return Dpred_k_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('Dpred k');
};
use constant::defer DpredInf_flat => sub {
  require MyFLAT;
  return DpredInf_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('DpredInf');
};
use constant::defer Spred_k_flat => sub {
  require MyFLAT;
  return Spred_k_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('Spred k');
};
use constant::defer SpredInf_flat => sub {
  require MyFLAT;
  return SpredInf_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('SpredInf');
};

use constant::defer DpredLeft_k_flat => sub {
  require MyFLAT;
  return Dpred_k_flat()->intersect(TurnLpred_flat())
    ->MyFLAT::set_name('DpredLeft k');
};
use constant::defer DpredRight_k_flat => sub {
  require MyFLAT;
  return Dpred_k_flat()->intersect(TurnRpred_flat())
    ->MyFLAT::set_name('DpredRight k');
};
use constant::defer SpredLeft_k_flat => sub {
  return Spred_k_flat()->intersect(TurnLpred_flat())
    ->MyFLAT::set_name('SpredLeft k');
};
use constant::defer SpredRight_k_flat => sub {
  return Spred_k_flat()->intersect(TurnRpred_flat())
    ->MyFLAT::set_name('SpredRight k');
};

{
  my @first_second_state_table = ([ 0,0,0,0,0 ],
                                  [ 1,3,2,5,4 ],   # 1 = start
                                  [ 3,2,6,6,2 ],   # 2 = lm0
                                  [ 3,3,2,7,7 ],   # 3 = lm1
                                  [ 6,6,5,4,4 ],   # 4 = rm0
                                  [ 5,7,7,5,4 ],   # 5 = rm1
                                  [ 6,6,6,6,6 ],   # 6 = DpredFirst
                                  [ 7,7,7,7,7 ]);  # 7 = DpredSecond

  use constant::defer DpredFirst_LtoH_flat => sub {
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA (\@first_second_state_table,
                                     name => 'DpredFirst LtoH',
                                     starting => 1,
                                     accepting => 6);
  };
  use constant::defer DpredSecond_LtoH_flat => sub {
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA (\@first_second_state_table,
                                     name => 'DpredSecond LtoH',
                                     starting => 1,
                                     accepting => 7);
  };
}
use constant::defer DpredFirst_flat => sub {
  require MyFLAT;
  return DpredFirst_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('DpredFirst')
    ->MyFLAT::minimize;
};
use constant::defer DpredSecond_flat => sub {
  require MyFLAT;
  return DpredSecond_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('DpredSecond')
    ->MyFLAT::minimize;
};


{
  my @Epred_LtoH_state_table = ([ 0,0,0,0,0 ],
                                [ 3,2,5,4,1 ],   # 1 = start
                                [ 3,2,6,6,2 ],   # 2 = lm0
                                [ 3,3,2,7,7 ],   # 3 = lm1
                                [ 6,6,5,4,4 ],   # 4 = rm0
                                [ 5,8,8,5,4 ],   # 5 = rm1
                                [ 6,6,6,6,6 ],   # 6 = non
                                [ 7,7,7,7,7 ],   # 7 = EpredR
                                [ 8,8,8,8,8 ]);  # 8 = EpredL

  use constant::defer EpredR_LtoH_flat => sub {
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA (\@Epred_LtoH_state_table,
                                     name => 'EpredR',
                                     starting => 1,
                                     accepting => 7);
  };
  use constant::defer EpredL_LtoH_flat => sub {
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA (\@Epred_LtoH_state_table,
                                     name => 'EpredL',
                                     starting => 1,
                                     accepting => 8);
  };
}
use constant::defer EpredR_flat => sub {
  require MyFLAT;
  return EpredR_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('EpredR')
    ->MyFLAT::minimize;
};
use constant::defer EpredL_flat => sub {
  require MyFLAT;
  return EpredL_LtoH_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('EpredL')
    ->MyFLAT::minimize;
};
use constant::defer Epred_flat => sub {
  require MyFLAT;
  return  EpredR_flat()->union(EpredL_flat())->as_dfa
    ->MyFLAT::set_name('Epred');
};

use constant::defer EpredR3_flat => sub {
  require MyFLAT;
  return  EpredR_flat()->MyFLAT::concat(FLAT::Regex->new('(0|4) 4'))->as_dfa
    ->MyFLAT::set_name('EpredR3');
};
use constant::defer EpredL3_flat => sub {
  require MyFLAT;
  return  EpredL_flat()->MyFLAT::concat(FLAT::Regex->new('2'))->as_dfa
    ->MyFLAT::set_name('EpredL3');
};

use constant::defer EpredSix_flat => sub {
  require MyFLAT;
  return  EpredL_flat()->MyFLAT::concat(FLAT::Regex->new('42'))->as_dfa
    ->MyFLAT::set_name('EpredSix');
};


#------------------------------------------------------------------------------

# 0,4 anywhere
# 3,1 in pairs
use constant::defer Opred_flat => sub {
  require MyFLAT;
  require FLAT::Regex;
  my $f = FLAT::Regex->new
    ('( (0|4)* 3 (0|4)* 1 )* (0|4)*') ->as_dfa;

  return $f->intersect(FLAT::Regex->new    # must be non-zero
                       ('(0|1|2|3|4)* (1|2|3|4) (0|1|2|3|4)*')->as_dfa)
    ->MyFLAT::set_name('Opred');
};

#------------------------------------------------------------------------------
{
  package MyR5DragonConf;
  use Carp 'croak';
  use List::Util 'min','max';
  use List::MoreUtils;

  sub new {
    my $class = shift;   # ($class, segs=>$s)
    if (ref $class) { $class = ref $class; }
    my $self = bless { @_ }, $class;
    if (exists $self->{'side'}) {
      croak 'MyR5DragonConf has no "side" parameter';
    }
    $self->{'segshash'} = { map {$_=>1} $self->segs_list };
    foreach my $seg ($self->segs_list) {
      $self->seg_validate($seg);
    }
    ### $self
    return $self;
  }
  sub clone {
    my $self = shift;   # ($class, segs=>$s)
    return $self->new (segs => [$self->segs_list],
                       @_);
  }
  sub seg_validate {
    my ($self, $seg) = @_;
    unless ($seg =~ /^(-?[1-9]|[0-9])[0-9]*,(-?[1-9]|[0-9])[0-9]*,[0123]$/) {
      croak "bad seg ",$seg," in ",$self->{'name'};
    }
    # x+y even goes horizontal dir=0,2, x+y odd goes vertical dir=1,3
    my ($x,$y,$dir) = split /,/, $seg;
    if (($x+$y + $dir) & 1) {
      croak "bad seg direction ",$seg," in ",$self->{'name'};
    }
  }
  sub key {
    my ($self) = @_;
    return ($self->{'key'} //= join(' ', sort $self->segs_list));
  }
  sub initial {
    my $class = shift;
    return $class->new (segs => ['0,0,0'], @_);
  }
  sub equals {
    my ($self, $other) = @_;
    return $self->key eq $other->key;
  }
  sub contains {
    my ($self, $seg) = @_;
    return $self->{'segshash'}->{$seg};
  }
  sub is_subset {
    my ($self, $superset) = @_;
    foreach my $seg ($self->segs_list) {
      if (! $superset->contains($seg)) {
        ### superset does not contain: $seg
        return 0;
      }
    }
    return 1;
  }
  sub segs_list {
    my ($self) = @_;
    return @{$self->{'segs'}};
  }
  sub num_segs {
    my ($self) = @_;
    return scalar @{$self->{'segs'}};
  }

  # return a new MyR5DragonConf with each point $x,$y moved to $x+$dx, $y+$dy
  sub translate {
    my ($self, $dx,$dy) = @_;
    if ($dx==0 && $dy==0) { return $self; }
    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name moved $dx,$dy";
    }
    return $self = $self->new
      (name => $name,
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       $x += $dx;
                       $y += $dy;
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return a new MyR5DragonConf with each segment rotated by $ddir, so each
  # $dir becomes $dir+$ddir, mod 4
  sub rotate {
    my ($self, $ddir) = @_;
    $ddir %= 4;
    if ($ddir == 0) { return $self; }

    my $name = $self->{'name'};
    if (defined $name) {
      $name = "$name rotate ".($ddir==1   ? "+90"
                               : $ddir==2 ? "180"
                               : $ddir==3 ? "-90"
                               : $ddir);
    }
    return $self = $self->new
      (name => $name,
       segs => [ map { my ($x,$y,$dir) = split /,/;
                       ($x,$y) = MyDragon::xy_rotate_plus90($x,$y, $ddir);
                       $dir = ($dir + $ddir) % 4;
                       "$x,$y,$dir"
                     } $self->segs_list ]);
  }

  # return a new MyR5DragonConf with the segment common to all the given confs
  sub intersection {
    my $self = shift;
    foreach my $other (@_) {
      $self = $self->new
        (within => $self->{'within'},
         segs => [ grep {$other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }
  sub difference {
    my $self = shift;
    foreach my $other (@_) {
      $self =$self->new
        (within => $self->{'within'},
         segs => [ grep {! $other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }
  sub union {
    my $self = shift;
    my $name = $self->{'name'};
    ### union(): $name
    if (defined $name) { $name = "$name union"; }
    return $self->new (name   => $name,
                       within => $self->{'within'},
                       segs   => [ List::MoreUtils::uniq
                                   (map {$_->segs_list} $self,@_) ]);

  }

  sub set_name {
    my ($self, $name) = @_;
    $self->{'name'} = $name;
    return $self;
  }
  sub set_within {
    my ($self, $within) = @_;
    $self->{'within'} = $within;
    return $self;
  }

  sub expand {
    my ($self, $digit) = @_;
    ### expand: $self->{'name'}
    ($digit>=0 && $digit<=4) or die;

    # state 0 right side enc, state 1 left side enc
    #  4 ---> 5
    #  ^
    #  |
    #  3 <--- 2
    #         ^
    #         |
    #  0 ---> 1

    my @segs = map { my ($x,$y,$dir) = split /,/;
                     ($x,$y) = MyR5Dragon::mul($x,$y);
                     my $x1 = $x + $dir4_to_dx[$dir];
                     my $y1 = $y + $dir4_to_dy[$dir];
                     my $dir1 = ($dir + 1) % 4;
                     my $x2 = $x1 + $dir4_to_dx[$dir1];
                     my $y2 = $y1 + $dir4_to_dy[$dir1];
                     my $x3 = $x + $dir4_to_dx[$dir1];
                     my $y3 = $y + $dir4_to_dy[$dir1];
                     my $x4 = $x3 + $dir4_to_dx[$dir1];
                     my $y4 = $y3 + $dir4_to_dy[$dir1];
                     my $dir2 = ($dir + 2) % 4;
                     ("$x,$y,$dir",
                      "$x1,$y1,$dir1",
                      "$x2,$y2,$dir2",
                      "$x3,$y3,$dir1",
                      "$x4,$y4,$dir")
                   } $self->segs_list;
    @segs = map { my ($x,$y,$dir) = split /,/;
                  $x -= $digit_to_x[$digit];
                  $y -= $digit_to_y[$digit];
                  foreach (1 .. $digit_to_dir[$digit]) {
                    ($x,$y) = MyR5Dragon::xy_rotate_minus90($x,$y);
                  }
                  $dir = ($dir - $digit_to_dir[$digit]) % 4;
                  "$x,$y,$dir"
                } @segs;
    my $expand = $self->new (segs => \@segs);
    if (defined (my $within = $self->{'within'})) {
      $expand = $expand->intersection($within);
      $expand->{'within'} = $within;
    }
    if (defined (my $name = $self->{'name'})) {
      unless ($name =~ /[0-9]$/) { $name .= ' '; }
      $expand->{'name'} = $name . '0';
    }
    return $expand;
  }

  # return true if the configuration has a boundary square side 0=right 1=left
  sub is_boundary {
    my ($self, $side) = @_;
    ### is_boundary(): $side
    my $within = $self->{'within'} // croak 'No within';
    my $dd = ($side & 1
              ?  1    # side=1 to left, dir+1
              : -1);  # side=0 to right, dir-1
    ### $dd

  SEG: foreach my $seg ($self->segs_list) {
      ### $seg
      my ($x,$y,$dir) = split /,/, $seg;

      #               dir
      #         x,y  -----> x1,y1
      #          ^          |
      #    dir3  |          | dir1
      #          |          v
      #       x3,y3 <----- x2,y2
      #              dir2
      my $dir1 = ($dir  + $dd) % 4;
      my $dir2 = ($dir1 + $dd) % 4;
      my $dir3 = ($dir2 + $dd) % 4;
      my $x1 = $x + $dir4_to_dx[$dir];
      my $y1 = $y + $dir4_to_dy[$dir];
      my $x2 = $x1 + $dir4_to_dx[$dir1];
      my $y2 = $y1 + $dir4_to_dy[$dir1];
      my $x3 = $x2 + $dir4_to_dx[$dir2];
      my $y3 = $y2 + $dir4_to_dy[$dir2];

      foreach my $other_seg ("$x1,$y1,$dir1",
                             "$x2,$y2,$dir2",
                             "$x2,$y2,$dir3") {
        ### $other_seg
        if (! $within->contains($other_seg)) {
          # presence of $other_seg is unspecified, this square cannot be
          # used to test for boundary
          ### not within ...
          next SEG;
        }
        if (! $self->contains($other_seg)) {
          ### absent, so boundary ...
          return 1;
        }
      }
    }
    return 0;
  }

  # bounding box around the segments
  sub bbox {
    my ($self) = @_;
    my $x_min=0;
    my $x_max=1;
    my $y_min=0;
    my $y_max=0;
    foreach my $seg ($self->segs_list) {
      my ($x,$y,$dir) = split /,/, $seg;
      foreach (0,1) {
        $x_min = min($x_min, $x); $x_max = max($x_max, $x);
        $y_min = min($y_min, $y); $y_max = max($y_max, $y);
        $x += $dir4_to_dx[$dir];
        $y += $dir4_to_dy[$dir];
      }
    }
    return ($x_min,$y_min, $x_max,$y_max);
  }
  sub print {
    my ($self) = @_;
    print $self->print_to_str, "\n";
  }
  sub print_to_str {
    my ($self) = @_;
    my $segshash = $self->{'segshash'};
    my $R = $self->{'within'} && $self->is_boundary(0);
    my $L = $self->{'within'} && $self->is_boundary(1);

    my $ret = "conf";
    if (defined $self->{'name'}) {
      $ret .= ' ' . $self->{'name'};
    }
    if ($self->{'within'}) {
      $ret .= ' boundary'; 
      $ret .= ($R ? " R" : " nonR");
      $ret .= ($L ? " L" : " nonL");
    }
    $ret .= "\n";

    # sort for consistency of output
    $ret .= "   " . join(' ',sort $self->segs_list) . "\n";

    my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
    ### $x_min
    ### $x_max
    ### $y_min
    ### $y_max

    foreach my $y (reverse $y_min .. $y_max) {
      my $yp1 = $y+1;
      if ($y<$y_max) {
        # above verticals
        $ret .= "   ";
        foreach my $x ($x_min .. $x_max) {
          if    ($self->contains("$x,$y,1"))   { $ret .= "^     "; }
          elsif ($self->contains("$x,$yp1,3")) { $ret .= "|     "; }
          else                                 { $ret .= "      "; }
        }
        $ret .= "\n";

        $ret .= "   ";
        foreach my $x ($x_min .. $x_max) {
          if    ($self->contains("$x,$y,1"))   { $ret .= "|     "; }
          elsif ($self->contains("$x,$yp1,3")) { $ret .= "|     "; }
          else                                 { $ret .= "      "; }
        }
        $ret .= "\n";

        $ret .= "   ";
        foreach my $x ($x_min .. $x_max) {
          if    ($self->contains("$x,$y,1"))   { $ret .= "|     "; }
          elsif ($self->contains("$x,$yp1,3")) { $ret .= "v     "; }
          else                                 { $ret .= "      "; }
        }
        $ret .= "\n";
      }

      # row
      $ret .= "   ";
      foreach my $x ($x_min .. $x_max) {
        my $xp1 = $x+1;
        $ret .= ($x==0&&$y==0 ? "O" : "*");
        if ($self->contains("$x,$y,0"))      { $ret .= ($x==0&&$y==0
                                                 ? "====>" : "---->"); }
        elsif ($self->contains("$xp1,$y,2")) { $ret .= "<----"; }
        else                                 { $ret .= "     "; }
      }
      $ret .= "\n";
    }
    $ret .= "\n";
    $ret =~ s/ +\n/\n/g;  # trailing whitespace
    return $ret;
  }
}

#------------------------------------------------------------------------------

#    0,0 ======> 1,0
#     ^           |
#     |           |
#     |           v
#    0,-1 <----- 1,-1
#
our $R_square_segs = MyR5DragonConf->new
  (name => 'R square',
   segs => [
            "0,0,0",       # below m
            "1,0,3",
            "1,-1,2",
            "0,-1,1",
           ]);

#    0,1 <------ 1,1
#     |           ^
#     |           |
#     v           |
#    0,0 ======> 1,0
#
our $L_square_segs = MyR5DragonConf->new
  (name => 'L square',
   segs => [
            "0,0,0",      # above m
            "1,0,1",
            "1,1,2",
            "0,1,3",
           ]);

#    0,1 <------ 1,1
#     ^           |
#     |           |
#     |           v
#    0,0 ======> 1,0
#     |           ^
#     |           |
#     v           |
#    0,-1 <----- 1,-1
#
our $B_squares_segs = $R_square_segs->union($L_square_segs)
  ->set_name('B two squares');  # Bpred enclosing


#------------------------------------------------------------------------------
# fBpred hullstouch

our $hullstouch_segs = MyR5DragonConf->new
  (name => 'hullstouch',
   segs => [
            '0,0,0',

            '0,0,2',   # left end
            '2,0,2',   # right end

            '1,0,1',   # left square
            '1,1,2',
            '0,1,3',

            '0,-1,1',  # right square
            '1,-1,2',
            '1,0,3',
           ]);
# $hullstouch_segs->print;

our @hullstouch_L_squares
  = ($L_square_segs, #  L top',
     MyR5DragonConf->new (name => 'L bottom left',
                          segs => [ '-1,-1,0',   
                                    '0,-1,1',
                                    '0,0,2',
                                    '-1,0,3' ]),
     MyR5DragonConf->new (name => 'L bottom right',
                          segs => [ '1,-1,0',   
                                    '2,-1,1',
                                    '2,0,2',
                                    '1,0,3' ]),
    );

our @hullstouch_R_squares
  = ($R_square_segs,
      MyR5DragonConf->new (name => 'R top left',
                           segs => [ '-1,1,0', 
                                     '0,1,3',
                                     '0,0,2',
                                     '-1,0,1' ]),
      MyR5DragonConf->new (name => 'R top right',
                           segs => [ '1,1,0', 
                                     '2,1,3',
                                     '2,0,2',
                                     '1,0,1' ]),
    );
# foreach my $conf (@hullstouch_L_squares,@hullstouch_R_squares){$conf->print;}

# $side==0 for R, $side==1 for L.
# A nearby square is boundary of respective $side when it has at least one
# segment $conf but not all of its segments in conf.
sub hullstouch_is_boundary {
  my ($conf, $side) = @_;
  foreach my $square ($side==0 ? @hullstouch_R_squares
                      : $side==1 ? @hullstouch_L_squares
                      : die) {
    my $got = $square->intersection ($conf);
    if ($got->num_segs == 0) {
      # no segments at all around this square, not boundary
      next;
    }
    my $all = $square->intersection ($conf->{'within'});
    if ($got->equals($all)) {
      # all segments (of "within") around this square, not boundary
      next;
    }
    # at least 1 but fewer than all segments, so boundary
    return 1;
  }
  return 0;
}

our (@fBpred_hullstouch_state_to_conf,
     @fBpred_hullstouch_state_table,@fBpred_hullstouch_state_enclosed_lists,
     $fNonRpred_hullstouch_flat,$fNonLpred_hullstouch_flat,
     $fNonBpred_hullstouch_flat,
     $fRpred_hullstouch_flat,$fLpred_hullstouch_flat,
     $fBpred_hullstouch_flat);
{
  my @state_is_RLB_enclosed;
  # my @state_is_RLB_enclosed_eventually;
  @fBpred_hullstouch_state_enclosed_lists = ([], [], []);

  use constant::defer fBpred_hullstouch_init => sub {
    my %options = @_;
    my $verbose = $options{'verbose'} || 0;
    ### fBpred_hullstouch_init() ...
    ### $verbose

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @fBpred_hullstouch_state_to_conf, $conf;
        $state = $#fBpred_hullstouch_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyR5DragonConf->initial(within => $hullstouch_segs));
    ### @fBpred_hullstouch_state_to_conf
    hullstouch_is_boundary($fBpred_hullstouch_state_to_conf[0],0) or die;
    hullstouch_is_boundary($fBpred_hullstouch_state_to_conf[0],1) or die;

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#fBpred_hullstouch_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $fBpred_hullstouch_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $digit (0 .. 4) {
            my $to_conf = $from_conf->expand($digit);
            my $to_state = $conf_to_state->($to_conf);
            $fBpred_hullstouch_state_table[$state]->[$digit] = $to_state;

            if ($verbose >= 3) {
              print "  digit=$digit to $fBpred_hullstouch_state_table[$state]->[$digit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#fBpred_hullstouch_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$fBpred_hullstouch_state_table[$from]->[0],1=>$fBpred_hullstouch_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $fBpred_hullstouch_state_table[$from]->[0] + 1;
      #   my $t1 = $fBpred_hullstouch_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@fBpred_hullstouch_state_table);
      print "fBpred hullstouch fBpred_hullstouch_state_table length $num_states\n";
    }

    # print "R_squares\n";
    # foreach my $segs (@R_squares) { segs_print($segs); }
    # print "L_squares\n";
    # foreach my $segs (@L_squares) { segs_print($segs); }

    foreach my $i (0 .. $#fBpred_hullstouch_state_to_conf) {
      my $conf = $fBpred_hullstouch_state_to_conf[$i];
      foreach my $side (0, 1) {
        $state_is_RLB_enclosed[$i]->[$side]
          = ! hullstouch_is_boundary($conf,$side);
      }
      $state_is_RLB_enclosed[$i]->[2]
        = $state_is_RLB_enclosed[$i]->[0]
        && $state_is_RLB_enclosed[$i]->[1];
    }
    foreach my $side (0, 1, 2) {
      $fBpred_hullstouch_state_enclosed_lists[$side]
        = [ grep {$state_is_RLB_enclosed[$_]->[$side]} 0 .. $#fBpred_hullstouch_state_table ];
    }
    if ($verbose >= 2) {
      print "  enclosed R list = ",join(',',@{$fBpred_hullstouch_state_enclosed_lists[0]}),
        " count ",scalar(@{$fBpred_hullstouch_state_enclosed_lists[0]}),"\n";
      print "  enclosed L list = ",join(',',@{$fBpred_hullstouch_state_enclosed_lists[1]}),
        " count ",scalar(@{$fBpred_hullstouch_state_enclosed_lists[1]}),"\n";
      print "  enclosed B list = ",join(',',@{$fBpred_hullstouch_state_enclosed_lists[2]}),
        " count ",scalar(@{$fBpred_hullstouch_state_enclosed_lists[2]}),"\n";
    }
    foreach my $side (2, 1, 0) {
      foreach my $i (@{$fBpred_hullstouch_state_enclosed_lists[$side]}) {
        foreach my $digit (0 .. 4) {
          my $to = $fBpred_hullstouch_state_table[$i]->[$digit];
          unless ($state_is_RLB_enclosed[$to]->[$side]) {
            $fBpred_hullstouch_state_to_conf[$i]->print;
            $fBpred_hullstouch_state_to_conf[$i]->expand($digit)->print;
            print "oops, side=$side enclosed $i digit=$digit didn't remain so";
            # print $conf->is_boundary($side);
            exit 1;
          }
        }
      }
    }

    # # eventually enclosed configurations = those with all of $segs_eventually_B
    # {
    #   my $bad = 0;
    #   my $side = 2;
    #   foreach my $state (0 .. $#fBpred_hullstouch_state_table) {
    #     my $conf = $fBpred_hullstouch_state_to_conf[$state];
    #     my $enc = $state_is_RLB_enclosed_eventually[$state]->[$side];
    #     my $allof = $segs_eventually_B->is_subset($conf) ? 1 : 0;
    #     if ($enc != $allof) {
    #       print "state $state enc $enc  triples $allof\n";
    #       conf_print($conf);
    #       $bad++;
    #     }
    #   }
    #   if ($bad) { exit 1; }
    # }

    require MyFLAT;
    {
      # R
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullstouch_state_table,
         name => 'fNonRpred_hullstouch',
         accepting_list => $fBpred_hullstouch_state_enclosed_lists[0],
        );
      $f = $f->MyFLAT::eventually_accepting;
      $fNonRpred_hullstouch_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }

      $f = $f->complement;
      $f->{'name'} = 'fRpred_hullstouch';
      $fRpred_hullstouch_flat = $f;
    }
    {
      # L
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullstouch_state_table, name => 'fNonLpred_hullstouch',
         starting => 0,
         accepting_list => $fBpred_hullstouch_state_enclosed_lists[1],
        );
      $f = $f->MyFLAT::eventually_accepting;
      $fNonLpred_hullstouch_flat = $f;

      $f = $f->complement;
      $f->{'name'} = 'fLpred_hullstouch';
      $fLpred_hullstouch_flat = $f;
    }
    {
      # B
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullstouch_state_table, name => 'fNonBpred_hullstouch',
         starting => 0,
         accepting_list => $fBpred_hullstouch_state_enclosed_lists[2],
        );
      $f = $f->MyFLAT::eventually_accepting;
      $fNonBpred_hullstouch_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }
      $f = $f->complement;
      $f->{'name'} = 'fBpred_hullstouch';
      $fBpred_hullstouch_flat = $f;
    }
  };
}


#------------------------------------------------------------------------------
# fBpred hulls around

our $hullsaround_segs_outer = MyR5DragonConf->new
  (name => 'hullsaround outer',
   segs => [
            "0,-2,0",  # bottom
            "0,-2,2",
            "2,-2,2",

            "2,-1,3",   # right
            "2,-1,1",
            "2,1,3",
            "2,1,1",

            "2,2,2",    # top
            "0,2,0",
            "0,2,2",

            "-1,2,3",   # left
            "-1,0,1",
            "-1,0,3",
            "-1,-2,1",
           ]);

our $hullsaround_segs_inner = MyR5DragonConf->new
  (name => 'hullsaround inner',
   segs => [
            "0,0,0",
            "0,0,2",
            "2,0,2",

            "1,0,1",
            "1,0,3",

            "0,1,1",
            "0,1,3",

            "0,-1,1",
            "0,-1,3",

            "-1,1,0",
            "1,1,0",
            "1,1,2",

            "-1,-1,0",
            "1,-1,0",
            "1,-1,2",


            "1,2,3",
            "1,-2,1",
           ]);

our $hullsaround_segs_all = $hullsaround_segs_outer->union($hullsaround_segs_inner)
  ->set_name('hullsaround all');


# # *     *<----*     *
# #       |     ^
# #       |     |
# #       v     |
# # *<----O====>*<----*
# #       ^     |
# #       |     |
# #       |     v
# # *     *<----*     *
# our $hullsaround_segs_eventually = MyR5DragonConf->new
#   (name => 'hullsaround eventually',
#    segs => [ '0,0,0',
#              '0,0,2',
#              '2,0,2',
#              '1,0,1',
#              '1,0,3',
#              '0,1,3',
#              '0,-1,1',
#              '1,1,2',
#              '1,-1,2',
#            ]);

our (@fBpred_hullsaround_state_to_conf,
     @fBpred_hullsaround_state_table,@fBpred_hullsaround_state_enclosed_lists,
     $fNonRpred_hullsaround_flat,$fNonLpred_hullsaround_flat,
     $fNonBpred_hullsaround_flat,
     $fRpred_hullsaround_flat,$fLpred_hullsaround_flat,
     $fBpred_hullsaround_flat);
{
  my @state_is_RLB_enclosed;
  # my @state_is_RLB_enclosed_eventually;
  @fBpred_hullsaround_state_enclosed_lists = ([], [], []);

  use constant::defer fBpred_hullsaround_init => sub {
    my %options = @_;
    my $verbose = $options{'verbose'} || 0;
    ### fBpred_hullsaround_init() ...
    ### $verbose

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @fBpred_hullsaround_state_to_conf, $conf;
        $state = $#fBpred_hullsaround_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyR5DragonConf->initial(within => $hullsaround_segs_all));
    ### @fBpred_hullsaround_state_to_conf
    $fBpred_hullsaround_state_to_conf[0]->is_boundary(0) or die;
    $fBpred_hullsaround_state_to_conf[0]->is_boundary(1) or die;

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#fBpred_hullsaround_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $fBpred_hullsaround_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $digit (0 .. 4) {
            my $to_conf = $from_conf->expand($digit);
            my $to_state = $conf_to_state->($to_conf);
            $fBpred_hullsaround_state_table[$state]->[$digit] = $to_state;

            if ($verbose >= 3) {
              print "  digit=$digit to $fBpred_hullsaround_state_table[$state]->[$digit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#fBpred_hullsaround_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$fBpred_hullsaround_state_table[$from]->[0],1=>$fBpred_hullsaround_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $fBpred_hullsaround_state_table[$from]->[0] + 1;
      #   my $t1 = $fBpred_hullsaround_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@fBpred_hullsaround_state_table);
      print "fBpred hullsaround fBpred_hullsaround_state_table length $num_states\n";
    }

    # print "R_squares\n";
    # foreach my $segs (@R_squares) { segs_print($segs); }
    # print "L_squares\n";
    # foreach my $segs (@L_squares) { segs_print($segs); }

    foreach my $i (0 .. $#fBpred_hullsaround_state_to_conf) {
      my $conf = $fBpred_hullsaround_state_to_conf[$i];
      foreach my $side (0, 1) {
        $state_is_RLB_enclosed[$i]->[$side]
          = ! $conf->is_boundary($side);
      }
      $state_is_RLB_enclosed[$i]->[2]
        = $state_is_RLB_enclosed[$i]->[0]
        && $state_is_RLB_enclosed[$i]->[1];
    }
    foreach my $side (0, 1, 2) {
      $fBpred_hullsaround_state_enclosed_lists[$side]
        = [ grep {$state_is_RLB_enclosed[$_]->[$side]} 0 .. $#fBpred_hullsaround_state_table ];
    }
    if ($verbose >= 2) {
      print "  enclosed R list = ",join(',',@{$fBpred_hullsaround_state_enclosed_lists[0]}),
        " count ",scalar(@{$fBpred_hullsaround_state_enclosed_lists[0]}),"\n";
      print "  enclosed L list = ",join(',',@{$fBpred_hullsaround_state_enclosed_lists[1]}),
        " count ",scalar(@{$fBpred_hullsaround_state_enclosed_lists[1]}),"\n";
      print "  enclosed B list = ",join(',',@{$fBpred_hullsaround_state_enclosed_lists[2]}),
        " count ",scalar(@{$fBpred_hullsaround_state_enclosed_lists[2]}),"\n";
    }
    foreach my $side (2, 1, 0) {
      foreach my $i (@{$fBpred_hullsaround_state_enclosed_lists[$side]}) {
        foreach my $digit (0 .. 4) {
          my $to = $fBpred_hullsaround_state_table[$i]->[$digit];
          unless ($state_is_RLB_enclosed[$to]->[$side]) {
            $fBpred_hullsaround_state_to_conf[$i]->print;
            $fBpred_hullsaround_state_to_conf[$i]->expand($digit)->print;
            print "oops, side=$side enclosed $i digit=$digit didn't remain so";
            # print $conf->is_boundary($side);
            exit 1;
          }
        }
      }
    }

    # # eventually enclosed configurations = those with all of $segs_eventually_B
    # {
    #   my $bad = 0;
    #   my $side = 2;
    #   foreach my $state (0 .. $#fBpred_hullsaround_state_table) {
    #     my $conf = $fBpred_hullsaround_state_to_conf[$state];
    #     my $enc = $state_is_RLB_enclosed_eventually[$state]->[$side];
    #     my $allof = $segs_eventually_B->is_subset($conf) ? 1 : 0;
    #     if ($enc != $allof) {
    #       print "state $state enc $enc  triples $allof\n";
    #       conf_print($conf);
    #       $bad++;
    #     }
    #   }
    #   if ($bad) { exit 1; }
    # }

    require MyFLAT;
    {
      # R
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullsaround_state_table,
         name => 'fNonRpred_hullsaround',
         accepting_list => $fBpred_hullsaround_state_enclosed_lists[0],
        );
      $f = $f->MyFLAT::eventually_accepting;
      $fNonRpred_hullsaround_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }

      $f = $f->complement;
      $f->{'name'} = 'fRpred_hullsaround';
      $fRpred_hullsaround_flat = $f;
    }
    {
      # L
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullsaround_state_table, name => 'fNonLpred_hullsaround',
         starting => 0,
         accepting_list => $fBpred_hullsaround_state_enclosed_lists[1],
        );
      $f = $f->MyFLAT::eventually_accepting;
      $fNonLpred_hullsaround_flat = $f;

      $f = $f->complement;
      $f->{'name'} = 'fLpred_hullsaround';
      $fLpred_hullsaround_flat = $f;
    }
    {
      # B
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullsaround_state_table, name => 'fNonBpred_hullsaround',
         starting => 0,
         accepting_list => $fBpred_hullsaround_state_enclosed_lists[2],
        );
      $f = $f->MyFLAT::eventually_accepting;
      $fNonBpred_hullsaround_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }
      $f = $f->complement;
      $f->{'name'} = 'fBpred_hullsaround';
      $fBpred_hullsaround_flat = $f;
    }
  };
}

#------------------------------------------------------------------------------
# Hwp, Hep

use constant Hwp_factor => pi/(2*atan2(2,1));

sub Hwp {
  my ($k) = @_;
  my $total_x = 0;
  my $total_y = 0;
  my $prev_j = -1;
  my $d = 1;
  for (;;) {
    my $j = $k - int(Hwp_factor() * $d);
    last if $j < 0;
    my ($x,$y) = bk($j);
    ($x,$y) = xy_rotate_plus90($x,$y, $d-1);
    $total_x += $x;
    $total_y += $y;
    $prev_j = $j;
    $d++;
  }
  # always term b^0 = 1, with $d factor
  if ($prev_j == 1) {
    my ($x,$y) = (1,0);
    ($x,$y) = xy_rotate_plus90($x,$y, $d-1);
    $total_x += $x;
    $total_y += $y;
  }
  return ($total_x, $total_y);
}

sub Hep {
  my ($k) = @_;
  my $total_x = 0;
  my $total_y = 0;
  my $prev_j = -1;
  my $d = 2;
  for (;;) {
    my $j = $k - int(Hwp_factor() * $d);
    last if $j < 0;
    my ($x,$y) = bk($j);
    ($x,$y) = xy_rotate_plus90($x,$y, $d-2);
    $total_x += $x;
    $total_y += $y;
    $prev_j = $j;
    $d++;
  }
  # always term b^0 = 1, with $d factor
  if ($prev_j == 1) {
    my ($x,$y) = (1,0);
    ($x,$y) = xy_rotate_plus90($x,$y, $d-2);
    $total_x += $x;
    $total_y += $y;
  }
  return ($total_x, $total_y);
}

use constant Hwpf_imag => -0.57698859;
use constant Hepf_real => -0.17698859;

use constant Hwpf_box_points => [ [    Hepf_real,  Hwpf_imag],
                                  [1 - Hepf_real,  Hwpf_imag],
                                  [1 - Hepf_real, -Hwpf_imag],
                                  [    Hepf_real, -Hwpf_imag] ];

use constant::defer Hwpf_k2_hull_points  => sub {
  my $hull_points = Hwpf_box_points();
  my $ex = 1;
  my $ey = 0;
  foreach (1 .. 2) {
    ($hull_points, $ex,$ey) = MyR5Dragon::hull_expand($hull_points, $ex,$ey);
  }
  my $z_end = Math::Complex->new($ex,$ey);
  my @z_points = map { Math::Complex->new(@$_) } @$hull_points;
  @z_points = map { $_ / $z_end } @z_points;
  return [ map {[$_->Re,$_->Im]} @z_points ];
};


#------------------------------------------------------------------------------
# Graph.pm

# sub Graph_xy_to_vertex {
#   my ($x,$y) = @_;
#   return sprintf '%d%+d', $x,$y;
# }
# sub Graph_vertex_to_xy {
#   my ($v) = @_;
#   $v =~ /(-?[0-9]+)\+?([-0-9]+)/ or die "bad vertex $v";
#   return ($1,$2);
# }

# return a Graph.pm graph of R5 dragon curve level $k
sub r5dragon_Graph {
  my ($k) = @_;
  require Graph;
  my $path = R5DragonCurve_path();
  my $graph = Graph->new (undirected => 1);
  $graph->set_graph_attribute (name => "R5 Dragon $k");
  $graph->set_graph_attribute('vertex_name_type','xy');
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  $graph->add_vertex("$prev_x,$prev_y");
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $graph->add_edge("$prev_x,$prev_y", "$x,$y");
    ($prev_x,$prev_y) = ($x,$y);
  }
  return $graph;
}

# return a Graph.pm graph of R5 quad area tree level $k
sub r5_area_tree_Graph {
  my ($n_start, $n_end) = @_;
  require Graph;
  my $path = R5DragonCurve_path();
  my $graph = Graph->new (undirected => 1);
  my ($x_start,$y_start) = $path->n_to_xy($n_start);
  my $prev;
  my $root;
  foreach my $n ($n_start .. $n_end-1) {
    my ($x,$y) = $path->n_to_xy($n);
    $x -= $x_start;
    $y -= $y_start;
    my ($dx,$dy) = $path->n_to_dxdy($n);
    $x += $dx/2; $y += $dy/2;
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left
    $x += $dx/2; $y += $dy/2;
    my $v = "$x,$y";
    if (! defined $root) { $root = $v; }
    $graph->add_vertex($v);
    if (defined $prev && $prev ne $v) { $graph->add_edge($v,$prev); }
    $prev = $v;
  }
  $graph->set_graph_attribute (root_vertex => $root);
  return $graph;
}
sub r5quad_area_tree_Graph {
  my ($k) = @_;
  my $n_start = 4*5**$k;
  my $n_end   = 8*5**$k;
  my $graph = r5_area_tree_Graph($n_start, $n_end);
  $graph->set_graph_attribute (name => "R5 Quad Area Tree $k");
  $graph->set_graph_attribute('vertex_name_type','xy');
  return $graph;
}
sub r5join_area_tree_Graph {
  my ($k) = @_;
  my $n_start = JN($k);
  my $n_end   = JNother($k);
  my $graph = r5_area_tree_Graph($n_start, $n_end);
  $graph->set_graph_attribute (name => "R5 Join Area Tree $k");
  $graph->set_graph_attribute('vertex_name_type','xy');
  return $graph;
}

sub QAdiampart_Graph {
  my ($n) = @_;
  ### QAdiampart_Graph(): $n
  require MyGraphs;
  my $k = 0;
  while ($n > 3**$k) { $k++; }
  my $graph = r5quad_area_tree_Graph($k);
  my @path = MyGraphs::Graph_tree_diameter_path($graph);
  @path >= $n or die;
  ### path length: scalar(@path)
  ### path start: $path[0]
  ### path end: $path[-1]
  if (defined (my $v = $path[$n])) {  # keep 0 .. $n-1
    $graph->delete_vertex($v);
    $graph->has_vertex($path[0]) or die;
    my $c = $graph->connected_component_by_vertex($path[0]);
    my @cc = $graph->connected_components;
    ### $c
    foreach my $i (0 .. $#cc) {
      if ($i != $c) {
        my @vertices = $graph->connected_component_by_index($i);
        ### delete: @vertices
        $graph->delete_vertices(@vertices);
      }
    }
  }
  $graph->set_graph_attribute (name => "R5 Quad Area Tree Diameter Part $n");
  return $graph;
}

# return a Graph.pm graph of R5 quad turns graph level $k
sub r5quad_turn_Graph {
  my ($k, %options) = @_;
  require Graph;
  my $path = R5DragonCurve_path();
  my $graph = Graph->new (undirected => 1);
  $graph->set_graph_attribute (name => "R5 Quad Turns $k");
  $graph->set_graph_attribute('vertex_name_type','xy');
  my $vertex_sequence = $options{'vertex_sequence'};
  my $n_start = 4*5**$k;
  my $n_end   = 8*5**$k;
  my ($x_start,$y_start) = $path->n_to_xy($n_start);
  my $prev;
  foreach my $n ($n_start .. $n_end-1) {
    my $turn = ($n==$n_start ? 1   # cycle rather than turning away
                : turn($n));
    if ($turn == -1) {
      my ($x,$y) = $path->n_to_xy($n);
      $x -= $x_start;
      $y -= $y_start;
      my $v = "$x,$y";
      ### $v
      if ($vertex_sequence && ! $graph->has_vertex($v)) {
        push @$vertex_sequence, $v;
      }
      $graph->add_vertex($v);
      if (defined $prev) { $graph->add_edge($v,$prev); }
      $prev = $v;
    } else {
      undef $prev;
    }
  }
  return $graph;
}

sub QAdiameter {
  my ($k) = @_;
  return 3**$k - 1;
}
sub QAdiameterCount {
  my ($k) = @_;
  return ($k==0 ? 1 : 6*9**($k-1));
}

sub JAdiameter {
  my ($k) = @_;
  if ($k == 0) { return 0; }
  return (3**$k - 3)/2;
}
sub JAdiameterCount {
  my ($k) = @_;
  return ($k==0 ? 1 : 3**($k-1));
}

sub QTdiameter {
  my ($k) = @_;
  if ($k == 0) { return 0; }
  return 3**$k + 2*$k - 3;
}

sub QAdomnum {
  my ($k) = @_;
  return ($k == 0 ? 1 : 5**($k-1));
}
sub JAdomnum {
  my ($k) = @_;
  return ($k == 0 ? 0 : (5**($k-1) + 3)/4);
}

#------------------------------------------------------------------------------

sub square_consecutive_type_xy {
  my ($x,$y, $n_start,$n_end) = @_;
  my $path = R5DragonCurve_path();
  my $dx = 1;
  my $dy = 0;
  my @list;
  foreach (1 .. 4) {
    my $x2 = $x + $dx;
    my $y2 = $y + $dy;
    my $n = $path->xyxy_to_n_either($x,$y, $x2,$y2);
    # ### side: "$x,$y  dxdy=$dx,$dy  to $x2,$y2  n=".($n//'undef')
    if (!defined $n) {
      ### no segment ...
      return '';
    }
    if ($n < $n_start || $n >= $n_end) {
      ### out of range ...
      return '';
    }
    push @list, $n;
    $x = $x2;
    $y = $y2;
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);
  }
  @list = map {
    my $from = $list[$_];
    my $to   = $list[$_-1];
    abs($from - $to) == 1
      || ($from==$n_start && $to  ==$n_end-1)
      || ($to  ==$n_start && $from==$n_end-1)
      ? 1 : 0
    } 0 .. $#list;

  # types 0000 0011 0110 0111
  if (sum(@list)) {
    until ($list[0]==0 && $list[-1]==1) {
      push @list, shift @list;
    }
  }
  my $str = join('',@list);
  if ($str eq '0011') {
    $str .= (($x+$y)%2 ? 'R' : 'L');
  }
  return $str;
}


#------------------------------------------------------------------------------

#       *-----*
#       |     |
#       /     |
# *----/./----*
#       / 4
#       |
#       |
#        -----
#             |
#             |
#             |
#       0-----
# of n, n+1 and n+2 are right turns
use constant::defer HQRpred_flat => sub {
  return TurnRpred_flat()->intersect
    (TurnLpred_flat()->MyFLAT::digits_increment(add => 1),
     TurnLpred_flat()->MyFLAT::digits_increment(add => 2),
     TurnLpred_flat()->MyFLAT::digits_increment(add => 3),
     TurnRpred_flat()->MyFLAT::digits_increment(add => 4),

     Rpred_flat()    ->MyFLAT::digits_increment(add => 0),
     Rpred_flat()    ->MyFLAT::digits_increment(add => 1),
     Rpred_flat()    ->MyFLAT::digits_increment(add => 2),
     Rpred_flat()    ->MyFLAT::digits_increment(add => 3))

    ->MyFLAT::minimize
    ->MyFLAT::set_name('HQRpred');
};

sub HQ {
  my ($k) = @_;
  $k>=0 or croak "HQ() is for k>=0";
  return ($k==0 ? 0 : 3**($k-1) - 1);
}
sub HQR {
  my ($k) = @_;
  $k>=0 or croak "HQR() is for k>=0";
  return HQ($k)/2;
}

#------------------------------------------------------------------------------

# Local variables:
# compile-command: "perl -w MyR5Dragon.pm"
# End:

1;
__END__
