#!/usr/bin/perl -w

# Copyright 2014, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use warnings;
use FindBin;
use File::stat ();
use File::Spec;
use File::Slurp;
use Image::Base::GD;
use IPC::Run;
use List::Util 'max','min';
use POSIX 'ceil';
use Math::PlanePath::R5DragonCurve 117; # v.117 for level_to_n_range()

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 1;    # or 2,2
my $scale = 4;

my $program_filename = File::Spec->catfile($FindBin::Bin,
                                           $FindBin::Script);
my $png_filename = File::Spec->catfile($FindBin::Bin,
                                       File::Spec->updir,
                                       'favicon.png');

my $path = Math::PlanePath::R5DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);
print "N= $n_lo to $n_hi\n";
my (@x,@y);
foreach my $n ($n_lo .. $n_hi) {
  my ($x,$y) = $path->n_to_xy($n);
  $x[$n] = $x;
  $y[$n] = $y;
}
my $x_min = min(@x);
my $y_min = min(@y);
my $x_max = max(@x);
my $y_max = max(@y);
### $x_min
### $y_min
### $x_max
### $y_max

my $w = $x_max - $x_min;
my $h = $y_max - $y_min;
print "X $x_min to $x_max = $w   Y $y_min to $y_max = $h\n";

my $size = 16;
my $width = $x_max - $x_min;
my $height = $y_max - $y_min;
my $x_base = int(($size - $width*$scale)/2 - $x_min*$scale);
my $y_base = int(($size - $height*$scale)/2 - $y_min*$scale);
### $width
### $height
### $x_base
### $y_base

my $image = Image::Base::GD->new (-height => $size,
                                  -width  => $size);
$image->rectangle(0,0, $size-1,$size-1, 'black');
foreach my $n ($n_lo .. $n_hi-1) {
  my $x1 = $x[$n];
  my $y1 = $y[$n];
  my $x2 = $x[$n+1];
  my $y2 = $y[$n+1];
  $x1 *= $scale;
  $y1 *= $scale;
  $x2 *= $scale;
  $y2 *= $scale;
  $x1 += $x_base;
  $y1 += $y_base;
  $x2 += $x_base;
  $y2 += $y_base;
  $y1 = $size-1 - $y1;
  $y2 = $size-1 - $y2;
  ### line: "$x1,$y1 to $x2,$y2"
  $image->line($x1,$y1, $x2,$y2, 'white');
}
$image->save($png_filename);

# PNG spec 11.3.4.2 suggests RFC822 (or rather RFC1123) for CreationTime
use constant STRFTIME_FORMAT_RFC822 => '%a, %d %b %Y %H:%M:%S %z';
my $program_fullname = File::Spec->catfile($FindBin::Bin, $FindBin::Script);
my $stat = File::stat::stat($program_fullname)
  // die "$program_fullname: $!";
my $program_mtime = $stat->mtime;
$ENV{'TZ'} = 'UTC+0';
my $creation_time = POSIX::strftime(STRFTIME_FORMAT_RFC822,
                                    gmtime($program_mtime));
### $creation_time

my $copyright = File::Slurp::read_file($program_filename);
$copyright =~ /^# (Copyright.*)/m or die;
$copyright = $1;

pngtextadd($png_filename, 'Author',    'Kevin Ryde');
pngtextadd($png_filename, 'Generator', "tools/$FindBin::Script");
pngtextadd($png_filename, 'Title',     'Iterations of the R5 Dragon Curve - icon');
pngtextadd($png_filename, 'Creation Time', $creation_time);
pngtextadd($png_filename, 'Copyright', $copyright);
pngtextadd($png_filename, 'Homepage',  'http://user42.tuxfamily.org/r5dragon/index.html');

print "uncompressed ",-s $png_filename,"\n";
IPC::Run::run(['optipng','-quiet','-o5',$png_filename]);
print "compressed   ",-s $png_filename,"\n";

# IPC::Run::run(['xzgv',$png_filename]);

exit 0;

# add text to the png image in $filename
sub pngtextadd {
  my ($filename, $keyword, $value) = @_;
  system('pngtextadd', "--keyword=$keyword", "--text=$value", $filename) == 0
    or die "system(pngtextadd)";
}
